import os
import numpy as np
from openpnm.topotools import trim, extend
from openpnm.utils import logging
from openpnm.io import GenericIO
from openpnm.network import GenericNetwork
from openpnm.geometry import GenericGeometry
import openpnm.models as mods
from pathlib import Path
from pandas import read_table, DataFrame
from tqdm import tqdm
logger = logging.getLogger(__name__)


class Statoil(GenericIO):
    r"""
    The StatOil format is used by the Maximal Ball network extraction code of
    the Imperial College London group

    This class can be used to load and work with those networks. Numerous
    datasets are available for download from the group's
    `website <http://tinyurl.com/zurko4q>`_.

    The 'Statoil' format consists of 4 different files in a single
    folder. The data is stored in columns with each corresponding to a
    specific property. Headers are not provided in the files, so one must
    refer to various theses and documents to interpret their meaning.
    """

    @classmethod
    def export_data(cls, network, shape, prefix=None, path=None, Pin=None, Pout=None):
        r"""

        Parameters
        ----------
        network : OpenPNM Network object
            The network
        shape : array_like
            An ndim-by-1 array or list containing the network dimensions
            in physical units (i.e. um)
        prefix : str
            The prefix to append to each file name, such as
            ``<prefix>_node1.dat``. If not provided ``project.name`` is used.
        path : str or path object
            The location where the exported files should be stored. If not
            provided the current working directory is used
        Pinlet and Poutlet : scalar, int (optional)
            The pore index of the inlet and outlet reservoir pores. If not
            provided then it is assumed they are the second last and last
            pores in the network, respectively.  This would be the case if
            the ``add_reservoir_pore`` function had been called prior to
            exporting.
        """
        if path is None:
            path = os.getcwd()
        p = Path(path)
        if prefix is None:
            prefix = network.project.name
        # Deal with reservoir pores
        if Pin is None:
            Pin = network.Np - 2
        if Pout is None:
            Pout = network.Np - 1
        Ptemp = network.find_neighbor_pores(pores=Pin)
        inlets = np.zeros_like(network.Ps, dtype=bool)
        inlets[Ptemp] = True
        Ptemp = network.find_neighbor_pores(pores=Pout)
        outlets = np.zeros_like(network.Ps, dtype=bool)
        outlets[Ptemp] = True

        # Write link 1 file
        props = ['throat.conns',
                 'throat.diameter',
                 'throat.shape_factor',
                 'throat.total_length']
        with open(p.joinpath(prefix + '_link1.dat'), 'wt') as f:
            f.write(str(network.Nt) + '\n')
            for row in tqdm(network.throats(), desc='Writing Link1 file'):
                s = ''
                s = s + '{:>9}'.format(str(row+1))
                for col in props:
                    try:
                        val = network[col][row]
                    except KeyError:
                        val = 0.0
                    if col == 'throat.conns':
                        val = np.copy(val)
                        val[val == network.Np - 1] = -1
                        val[val == (network.Np - 2)] = -2
                        s = s + '{:>9}'.format(str(val[0] + 1))
                        s = s + '{:>9}'.format(str(val[1] + 1))
                        continue
                    if isinstance(val, float):
                        if 'diameter' in col:  # Convert to radius
                            val = val/2
                        if np.isnan(val):
                            val = 0.0
                        val = np.format_float_scientific(val, precision=6,
                                                         exp_digits=3,
                                                         trim='k',
                                                         unique=False)
                    s = s + '{:>15}'.format(str(val))
                s = s + '\n'  # Remove trailing tab and add a new line
                f.write(s)

        # Write Link 2 file
        props = ['throat.conns',
                 'throat.conduit_lengths.pore1',
                 'throat.conduit_lengths.pore2',
                 'throat.conduit_lengths.throat',
                 'throat.volume',
                 'throat.clay_volume']
        with open(p.joinpath(prefix + '_link2.dat'), 'wt') as f:
            for row in tqdm(network.throats(), desc='Writing Link2 file'):
                s = ''
                s = s + '{:>9}'.format(str(row+1))
                for col in props:
                    try:
                        val = network[col][row]
                    except KeyError:
                        val = 0.0
                    if col == 'throat.conns':
                        val = np.copy(val)
                        val[val == network.Np - 1] = -1
                        val[val == (network.Np - 2)] = -2
                        # Original file has 7 spaces for pore indices, but
                        # this is not enough for networks with > 10 million
                        # pores so I have bumped it to 9. I'm not sure if
                        # this will still work with the ICL binaries.
                        s = s + '{:>9}'.format(str(val[0] + 1))
                        s = s + '{:>9}'.format(str(val[1] + 1))
                        continue
                    if isinstance(val, float):
                        if np.isnan(val):
                            val = 0.0
                        val = np.format_float_scientific(val, precision=6,
                                                         exp_digits=3,
                                                         trim='k',
                                                         unique=False)
                    s = s + '{:>15}'.format(str(val))
                s = s + '\n'  # Remove trailing tab and a new line
                f.write(s)

        # Write Node 1 file
        with open(p.joinpath(prefix + '_node1.dat'), 'wt') as f:
            s = ''
            s = s + str(network.num_pores('reservoir', 'not'))
            for d in shape:
                val = np.format_float_scientific(d, precision=6, exp_digits=3,
                                                 trim='k', unique=False)
                s = s + '{:>17}'.format(str(val))
            s = s + '\n'
            f.write(s)
            for row in tqdm(network.pores('reservoir', mode='not'),
                            desc='Writing Node1 file'):
                if row in [Pin, Pout]:
                    continue
                s = ''
                s = s + '{:>9}'.format(str(row+1))
                for c in network['pore.coords'][row]:
                    if isinstance(c, float):
                        c = np.format_float_scientific(c, precision=6,
                                                       exp_digits=3,
                                                       trim='k',
                                                       unique=False)
                    s = s + '{:>15}'.format(str(c))
                s = s + '{:>9}'.format(str(network.num_neighbors(row)[0]))
                for n in network.find_neighbor_pores(row):
                    if n == Pin:
                        n = 0
                    elif n == Pout:
                        n = -1
                    else:
                        n = n + 1
                    s = s + '{:>9}'.format(str(n))
                s = s + '{:>9}'.format(str(int(inlets[row])))
                s = s + '{:>9}'.format(str(int(outlets[row])))
                for n in network.find_neighbor_throats(row):
                    s = s + '{:>9}'.format(str(n + 1))
                s = s + '\n'  # Remove trailing tab and a new line
                f.write(s)

        # Write Node 2 file
        props = ['pore.volume',
                 'pore.diameter',
                 'pore.shape_factor',
                 'pore.clay_volume']
        with open(p.joinpath(prefix + '_node2.dat'), 'wt') as f:
            for row in tqdm(network.pores('reservoir', mode='not'),
                            desc='Writing Node2 file'):
                s = ''
                s = s + '{:>9}'.format(str(row+1))
                for col in props:
                    try:
                        val = network[col][row]
                    except KeyError:
                        val = 0.0
                    if isinstance(val, float):
                        if 'diameter' in col:
                            val = val/2
                        val = np.format_float_scientific(val, precision=6,
                                                         exp_digits=3,
                                                         trim='k',
                                                         unique=False)
                        s = s + '{:>15}'.format(str(val))
                s = s + '\n'  # Remove trailing tab and a new line
                f.write(s)

    @classmethod
    def load(cls, *args, **kwargs):
        r"""
        This method is being deprecated.  Use ``import_data`` instead.
        """
        return cls.import_data(*args, **kwargs)

    @classmethod
    def import_data(cls, path, prefix, network=None):
        r"""
        Load data from the \'dat\' files located in specified folder.

        Parameters
        ----------
        path : string
            The full path to the folder containing the set of \'dat\' files.

        prefix : string
            The file name prefix on each file. The data files are stored
            as \<prefix\>_node1.dat.

        network : OpenPNM Network Object
            If given then the data will be loaded on it and returned.  If not
            given, a Network will be created and returned.

        Returns
        -------
        An OpenPNM Project containing a GenericNetwork holding all the data

        """
        net = {}

        # Parse the link1 file
        path = Path(path)
        filename = Path(path.resolve(), prefix+'_link1.dat')
        with open(filename, mode='r') as f:
            link1 = read_table(filepath_or_buffer=f,
                               header=None,
                               skiprows=1,
                               sep=' ',
                               skipinitialspace=True,
                               index_col=0)
        link1.columns = ['throat.pore1', 'throat.pore2', 'throat.radius',
                         'throat.shape_factor', 'throat.total_length']
        # Add link1 props to net
        net['throat.conns'] = np.vstack((link1['throat.pore1']-1,
                                         link1['throat.pore2']-1)).T
        net['throat.conns'] = np.sort(net['throat.conns'], axis=1)
        net['throat.radius'] = np.array(link1['throat.radius'])
        net['throat.shape_factor'] = np.array(link1['throat.shape_factor'])
        net['throat.total_length'] = np.array(link1['throat.total_length'])

        filename = Path(path.resolve(), prefix+'_link2.dat')
        with open(filename, mode='r') as f:
            link2 = read_table(filepath_or_buffer=f,
                               header=None,
                               sep=' ',
                               skipinitialspace=True,
                               index_col=0)
        link2.columns = ['throat.pore1', 'throat.pore2',
                         'throat.pore1_length', 'throat.pore2_length',
                         'throat.length', 'throat.volume',
                         'throat.clay_volume']
        # Add link2 props to net
        cl_t = np.array(link2['throat.length'])
        net['throat.length'] = cl_t
        net['throat.conduit_lengths.throat'] = cl_t
        net['throat.volume'] = np.array(link2['throat.volume'])
        cl_p1 = np.array(link2['throat.pore1_length'])
        net['throat.conduit_lengths.pore1'] = cl_p1
        cl_p2 = np.array(link2['throat.pore2_length'])
        net['throat.conduit_lengths.pore2'] = cl_p2
        net['throat.clay_volume'] = np.array(link2['throat.clay_volume'])
        # ---------------------------------------------------------------------
        # Parse the node1 file
        filename = Path(path.resolve(), prefix+'_node1.dat')
        with open(filename, mode='r') as f:
            row_0 = f.readline().split()
            num_lines = int(row_0[0])
            array = np.ndarray([num_lines, 6])
            for i in range(num_lines):
                row = f.readline()\
                       .replace('\t', ' ').replace('\n', ' ').split()
                array[i, :] = row[0:6]
        node1 = DataFrame(array[:, [1, 2, 3, 4]])
        node1.columns = ['pore.x_coord', 'pore.y_coord', 'pore.z_coord',
                         'pore.coordination_number']
        # Add node1 props to net
        net['pore.coords'] = np.vstack((node1['pore.x_coord'],
                                        node1['pore.y_coord'],
                                        node1['pore.z_coord'])).T
        # ---------------------------------------------------------------------
        # Parse the node2 file
        filename = Path(path.resolve(), prefix+'_node2.dat')
        with open(filename, mode='r') as f:
            node2 = read_table(filepath_or_buffer=f,
                               header=None,
                               sep=' ',
                               skipinitialspace=True,
                               index_col=0)
        node2.columns = ['pore.volume', 'pore.radius', 'pore.shape_factor',
                         'pore.clay_volume']
        # Add node2 props to net
        net['pore.volume'] = np.array(node2['pore.volume'])
        net['pore.radius'] = np.array(node2['pore.radius'])
        net['pore.shape_factor'] = np.array(node2['pore.shape_factor'])
        net['pore.clay_volume'] = np.array(node2['pore.clay_volume'])
        net['throat.area'] = ((net['throat.radius']**2)
                              / (4.0*net['throat.shape_factor']))
        net['pore.area'] = ((net['pore.radius']**2)
                            / (4.0*net['pore.shape_factor']))

        if network is None:
            network = GenericNetwork()
        network = cls._update_network(network=network, net=net)

        # Use OpenPNM Tools to clean up network
        # Trim throats connected to 'inlet' or 'outlet' reservoirs
        trim1 = np.where(np.any(net['throat.conns'] == -1, axis=1))[0]
        # Apply 'outlet' label to these pores
        outlets = network['throat.conns'][trim1, 1]
        network['pore.outlets'] = False
        network['pore.outlets'][outlets] = True
        trim2 = np.where(np.any(net['throat.conns'] == -2, axis=1))[0]
        # Apply 'inlet' label to these pores
        inlets = network['throat.conns'][trim2, 1]
        network['pore.inlets'] = False
        network['pore.inlets'][inlets] = True
        # Now trim the throats
        to_trim = np.hstack([trim1, trim2])
        trim(network=network, throats=to_trim)

        return network.project

    @classmethod
    def add_reservoir_pore(cls, network, pores, offset=0.1):
        r"""

        Parameters
        ----------
        network : OpenPNM Network object
            The network to which the reservoir pore should be added
        pores : array_like
            The pores to which the reservoir pore should be connected to
        offset : scalar
            Controls the distance which the reservoir is offset from the given
            ``pores``.  The total displacement is found from the network
            dimension normal to given ``pores``, multiplied by ``offset``.
        """

        # Check if a label was given and fetch actual indices
        if isinstance(pores, str):
            # Convert 'face' into 'pore.face' if necessary
            if not pores.startswith('pore.'):
                pores = 'pore.' + pores
            pores = network.pores(pores)
        # Find coordinates of pores on given face
        coords = network['pore.coords'][pores]
        # Normalize the coordinates based on full network size
        c_norm = coords/network['pore.coords'].max(axis=0)
        # Identify axis of face by looking for dim with smallest delta
        diffs = np.amax(c_norm - np.average(c_norm, axis=0), axis=0)
        ax = np.where(diffs == diffs.min())[0][0]
        # Add new pore at center of domain
        new_coord = network['pore.coords'].mean(axis=0)
        domain_half_length = np.ptp(network['pore.coords'][:, ax])/2
        if coords[:, ax].mean() < network['pore.coords'][:, ax].mean():
            new_coord[ax] = new_coord[ax] - domain_half_length*(1 + offset)
        if coords[:, ax].mean() > network['pore.coords'][:, ax].mean():
            new_coord[ax] = new_coord[ax] + domain_half_length*(1 + offset)
        Ps = np.arange(network.Np, network.Np + 1)
        extend(network=network, coords=[new_coord], labels=['reservoir'])
        conns = [[P, network.Np-1] for P in pores]
        Ts = np.arange(network.Nt, network.Nt + len(conns))
        extend(network=network, conns=conns, labels=['reservoir'])
        # Compute the geometrical properties of the reservoir pore and throats
        # Confirm if network has any geometry props on it
        props = {'throat.length', 'pore.diameter', 'throat.volume'}
        if len(set(network.keys()).intersection(props)) > 0:
            raise Exception('Geometrical properties should be moved to a ' +
                            'geometry object first')
            # or just do this?:  geo = Imported(network=network)
        geo = GenericGeometry(network=network, pores=Ps, throats=Ts)
        geo.add_model(propname='pore.diameter',
                      model=mods.geometry.pore_size.largest_sphere)
        geo.add_model(propname='throat.diameter_temp',
                      model=mods.geometry.throat_size.from_neighbor_pores,
                      mode='min')
        geo.add_model(propname='throat.diameter',
                      model=mods.misc.scaled,
                      prop='throat.diameter_temp', factor=0.5)
        geo.add_model(propname='throat.length',
                      model=mods.geometry.throat_length.classic)
        geo.add_model(propname='throat.endpoints',
                      model=mods.geometry.throat_endpoints.spherical_pores)
        geo.add_model(propname='throat.conduit_lengths',
                      model=mods.geometry.throat_length.conduit_lengths)
        geo.add_model(propname='throat.volume',
                      model=mods.geometry.throat_volume.cylinder)


def get_domain_shape(network, pore_diameter='pore.diameter'):
    xmin, ymin, zmin = np.amin(network['pore.coords'], axis=0)
    xmax, ymax, zmax = np.amax(network['pore.coords'], axis=0)
    mins = []
    for axis, val in enumerate([xmin, ymin, zmin]):
        inds = np.where(network['pore.coords'][:, axis] == val)
        Rp = np.amax(network[pore_diameter][inds])/2
        mins.append(val - max(0, Rp))
    maxes = []
    for axis, val in enumerate([xmax, ymax, zmax]):
        inds = np.where(network['pore.coords'][:, axis] == val)
        Rp = np.amax(network[pore_diameter][inds])/2
        maxes.append(val + max(0, Rp))
    shape = np.array(maxes) - np.array(mins)
    return shape



from .GenericMixture import GenericMixture, LiquidMixture, GasMixture
from .GenericSpecies import GenericSpecies
from .SpeciesByName import SpeciesByName, LiquidByName, GasByName
from . import species
from .IdealGas import IdealGas
from .SalineWater import SalineWater
from .DryAir import DryAir
from .HumidAir import HumidAir



from openpnm.physics import GenericPhysics
from openpnm.models import physics as mods

__all__ = ["Basic"]


class Basic(GenericPhysics):
    r"""
    Minimal subclass of GenericPhysics for performing diffusion and/or
    flow simulations.

    Parameters
    ----------
    network : GenericNetwork
        The network to which this Physics should be attached/

    phase : GenericPhase
        The Phase object to which this Physics applies.

    geometry : GenericGeometry
        The Geometry object that defines the pores/throats where this
        Physics should be applied.

    name : str, optional
        A unique string name to identify the Physics object, typically
        same as instance name but can be anything.  If left blank, and
        name will be generated that include the class name and a random
        string.

    Returns
    -------
    GenericPhysics
        Basic subclass of GenericPhysics with minimal models attached
        required to perform diffusion and/or flow simulations.

    """

    def __init__(self, project=None, network=None, phase=None,
                 geometry=None, settings={}, **kwargs):

        # Define some default settings
        self.settings.update({'prefix': 'phys'})
        # Overwrite with user supplied settings, if any
        self.settings.update(settings)

        # Deal with network or project arguments
        if network is not None:
            if project is not None:
                assert network is project.network
            else:
                project = network.project

        super().__init__(project=project, phase=phase, geometry=geometry,
                         **kwargs)

        self.add_model(propname='throat.hydraulic_conductance',
                       model=mods.hydraulic_conductance.generic_hydraulic)
        self.add_model(propname='throat.diffusive_conductance',
                       model=mods.diffusive_conductance.generic_diffusive)
        self.add_model(propname='throat.entry_pressure',
                       model=mods.capillary_pressure.washburn)



import numpy as _np
import openpnm.geometry.GenericGeometry as _GenericGeometry
from .misc import _get_conduit_diameters
from .throat_length import ctc as _ctc

__all__ = [
    "spheres_and_cylinders",
    "circles_and_rectangles",
    "cones_and_cylinders",
    "trapezoids_and_rectangles",
    "pyramids_and_cuboids",
    "cubes_and_cuboids",
    "squares_and_rectangles"
]


def spheres_and_cylinders(
    target: _GenericGeometry,
    pore_diameter="pore.diameter",
    throat_diameter="throat.diameter",
):
    r"""
    Calculates conduit lengths in the network assuming pores are spheres
    and throats are cylinders.

    A conduit is defined as ( 1/2 pore - full throat - 1/2 pore ).

    Parameters
    ----------
    target : GenericGeometry
        Geometry object which this model is associated with. This controls
        the length of the calculated array, and also provides access to
        other necessary properties.
    pore_diameter : str
        Dictionary key of the pore diameter values.
    throat_diameter : str
        Dictionary key of the throat diameter values.

    Returns
    -------
    ndarray
        Array (Nt by 3) containing conduit lengths' values for conduits in
        the given geometry object. The array is formatted as:
            [pore1, throat, pore2].T

    """
    L_ctc = _ctc(target)
    D1, Dt, D2 = _get_conduit_diameters(target, pore_diameter, throat_diameter)

    # Handle the case where Dt > Dp
    if (Dt > D1).any() or (Dt > D2).any():
        _raise_incompatible_data()
    L1 = _np.sqrt(D1**2 - Dt**2) / 2
    L2 = _np.sqrt(D2**2 - Dt**2) / 2

    # Handle throats w/ overlapping pores
    _L1 = (4 * L_ctc**2 + D1**2 - D2**2) / (8 * L_ctc)
    mask = L_ctc - 0.5 * (D1 + D2) < 0
    L1[mask] = _L1[mask]
    L2[mask] = (L_ctc - L1)[mask]

    Lt = _np.maximum(L_ctc - (L1 + L2), 1e-15)

    return _np.vstack((L1, Lt, L2)).T


def circles_and_rectangles(
    target,
    pore_diameter="pore.diameter",
    throat_diameter="throat.diameter"
):
    r"""
    Calculates conduit lengths in the network assuming pores are circles
    and throats are rectangles.

    A conduit is defined as ( 1/2 pore - full throat - 1/2 pore ).

    Parameters
    ----------
    target : GenericGeometry
        Geometry object which this model is associated with. This controls
        the length of the calculated array, and also provides access to
        other necessary properties.
    pore_diameter : str
        Dictionary key of the pore diameter values.
    throat_diameter : str
        Dictionary key of the throat diameter values.

    Returns
    -------
    ndarray
        Array (Nt by 3) containing conduit lengths' values for conduits in
        the given geometry object. The array is formatted as:
            [pore1, throat, pore2].T

    Notes
    -----
    This model should only be used for true 2D networks, i.e. with planar
    symmetry.

    """
    return spheres_and_cylinders(
        target=target,
        pore_diameter=pore_diameter,
        throat_diameter=throat_diameter
    )


def cones_and_cylinders(
    target,
    pore_diameter="pore.diameter",
    throat_diameter="throat.diameter"
):
    r"""
    Calculates conduit lengths in the network assuming pores are cones
    and throats are cylinders.

    A conduit is defined as ( 1/2 pore - full throat - 1/2 pore ).

    Parameters
    ----------
    target : GenericGeometry
        Geometry object which this model is associated with. This controls
        the length of the calculated array, and also provides access to
        other necessary properties.
    pore_diameter : str
        Dictionary key of the pore diameter values.
    throat_diameter : str
        Dictionary key of the throat diameter values.

    Returns
    -------
    ndarray
        Array (Nt by 3) containing conduit lengths' values for conduits in
        the given geometry object. The array is formatted as:
            [pore1, throat, pore2].T

    """
    L_ctc = _ctc(target)
    D1, Dt, D2 = _get_conduit_diameters(target, pore_diameter, throat_diameter)

    L1 = D1 / 2
    L2 = D2 / 2

    # Handle throats w/ overlapping pores
    _L1 = (4 * L_ctc**2 + D1**2 - D2**2) / (8 * L_ctc)
    mask = L_ctc - 0.5 * (D1 + D2) < 0
    L1[mask] = _L1[mask]
    L2[mask] = (L_ctc - L1)[mask]

    Lt = _np.maximum(L_ctc - (L1 + L2), 1e-15)

    return _np.vstack((L1, Lt, L2)).T


def trapezoids_and_rectangles(
    target,
    pore_diameter="pore.diameter",
    throat_diameter="throat.diameter"
):
    r"""
    Calculates conduit lengths in the network assuming pores are
    trapezoids and throats are rectangles.

    A conduit is defined as ( 1/2 pore - full throat - 1/2 pore ).

    Parameters
    ----------
    target : GenericGeometry
        Geometry object which this model is associated with. This controls
        the length of the calculated array, and also provides access to
        other necessary properties.
    pore_diameter : str
        Dictionary key of the pore diameter values.
    throat_diameter : str
        Dictionary key of the throat diameter values.

    Returns
    -------
    ndarray
        Array (Nt by 3) containing conduit lengths' values for conduits in
        the given geometry object. The array is formatted as:
            [pore1, throat, pore2].T

    Notes
    -----
    This model should only be used for true 2D networks, i.e. with planar
    symmetry.

    """
    L_ctc = _ctc(target)
    D1, Dt, D2 = _get_conduit_diameters(target, pore_diameter, throat_diameter)

    L1 = D1 / 2
    L2 = D2 / 2

    # Handle throats w/ overlapping pores
    _L1 = (4 * L_ctc**2 + D1**2 - D2**2) / (8 * L_ctc)
    mask = L_ctc - 0.5 * (D1 + D2) < 0
    L1[mask] = _L1[mask]
    L2[mask] = (L_ctc - L1)[mask]

    Lt = _np.maximum(L_ctc - (L1 + L2), 1e-15)

    return _np.vstack((L1, Lt, L2)).T


def pyramids_and_cuboids(
    target,
    pore_diameter="pore.diameter",
    throat_diameter="throat.diameter"
):
    r"""
    Calculates conduit lengths in the network assuming pores are truncated
    pyramids and throats are cuboids.

    A conduit is defined as ( 1/2 pore - full throat - 1/2 pore ).

    Parameters
    ----------
    target : GenericGeometry
        Geometry object which this model is associated with. This controls
        the length of the calculated array, and also provides access to
        other necessary properties.
    pore_diameter : str
        Dictionary key of the pore diameter values.
    throat_diameter : str
        Dictionary key of the throat diameter values.

    Returns
    -------
    ndarray
        Array (Nt by 3) containing conduit lengths' values for conduits in
        the given geometry object. The array is formatted as:
            [pore1, throat, pore2].T

    """
    return cones_and_cylinders(
        target, pore_diameter=pore_diameter, throat_diameter=throat_diameter
    )


def cubes_and_cuboids(
    target,
    pore_diameter="pore.diameter",
    throat_diameter="throat.diameter"
):
    r"""
    Calculates conduit lengths in the network assuming pores are cubes
    and throats are cuboids.

    A conduit is defined as ( 1/2 pore - full throat - 1/2 pore ).

    Parameters
    ----------
    target : GenericGeometry
        Geometry object which this model is associated with. This controls
        the length of the calculated array, and also provides access to
        other necessary properties.
    pore_diameter : str
        Dictionary key of the pore diameter values.
    throat_diameter : str
        Dictionary key of the throat diameter values.

    Returns
    -------
    ndarray
        Array (Nt by 3) containing conduit lengths' values for conduits in
        the given geometry object. The array is formatted as:
            [pore1, throat, pore2].T

    """
    L_ctc = _ctc(target)
    D1, Dt, D2 = _get_conduit_diameters(target, pore_diameter, throat_diameter)

    L1 = D1 / 2
    L2 = D2 / 2
    Lt = L_ctc - (L1 + L2)

    # Handle overlapping pores
    mask = (Lt < 0) & (L1 > L2)
    L2[mask] = (L_ctc - L1)[mask]
    mask = (Lt < 0) & (L2 > L1)
    L1[mask] = (L_ctc - L2)[mask]
    Lt = _np.maximum(Lt, 1e-15)

    return _np.vstack((L1, Lt, L2)).T


def squares_and_rectangles(
    target,
    pore_diameter="pore.diameter",
    throat_diameter="throat.diameter"
):
    r"""
    Calculates conduit lengths in the network assuming pores are squares
    and throats are rectangles.

    A conduit is defined as ( 1/2 pore - full throat - 1/2 pore ).

    Parameters
    ----------
    target : GenericGeometry
        Geometry object which this model is associated with. This controls
        the length of the calculated array, and also provides access to
        other necessary properties.
    pore_diameter : str
        Dictionary key of the pore diameter values.
    throat_diameter : str
        Dictionary key of the throat diameter values.

    Returns
    -------
    ndarray
        Array (Nt by 3) containing conduit lengths' values for conduits in
        the given geometry object. The array is formatted as:
            [pore1, throat, pore2].T

    Notes
    -----
    This model should only be used for true 2D networks, i.e. with planar
    symmetry.

    """
    return cubes_and_cuboids(
        target, pore_diameter=pore_diameter, throat_diameter=throat_diameter
    )


# Dealing with errors and exceptions
def _raise_incompatible_data():
    raise Exception(
        "'spheres_and_cylinders' can only be applied when throat diameter is"
        " smaller than that of adjacent pores."
    )



import subprocess
from pathlib import Path


class PNFlow:
    r"""
    run pnflow.exe file on a network

    Notes
    -----
    This function only runs on Windows since the Windows compatible binary is
    provided by the Imperial College team.
    """
    @classmethod
    def run(cls, prefix, path='./'):
        r"""
        run pnflow.exe on a network

        Parameters
        ----------
        prefix : string
            The prefix of the filenames (i.e. 'prefix_node1.dat')

        Notes
        -----
        All the required files including prefix_link1, prefix_link2,
        prefix_node1, prefix_node2, prefix.dat and pnflow.exe file
        should be in a same directory as the exectuable.
        """

        # Create dat file
        s = f"TITLE  {prefix}; \n" \
            f"writeStatistics true; \n" \
            f"NETWORK  F {prefix} \n" \
            f"cycle1:   0.00     2.00E+05     0.05      T     T; \n" \
            f"cycle2:   1.00    -2.00E+05     0.05      T     T; \n" \
            f"cycle3:   0.00     2.00E+05     0.05      T     T; \n" \
            f"cycle1_BC:   T     F       T       T      DP    1.00  2.00; \n" \
            f"cycle2_BC:   T     F       T       T      DP    2.00  1.00; \n" \
            f"cycle3_BC:   T     F       T       T      DP    1.00  2.00; \n" \
            f"CALC_BOX:  0.1 0.9; \n" \
            f"INIT_CONT_ANG:   1   0   10  -0.2    -3.0   rand   0.0; \n" \
            f"EQUIL_CON_ANG:   4   30   50   -0.2   -3.0   rand   0.0; \n" \
            f"Water  0.001               1.2                1000.0; \n" \
            f"Oil   0.001               1000.0              1000.0; \n" \
            f"ClayResistivity            2.0 ; \n" \
            f"WaterOilInterface          0.03 ; \n" \
            f"DRAIN_SINGLETS: T; \n" \
            f"SAT_CONTROL: F; \n"

        file = Path(path + prefix + '.dat')
        with open(file, 'w') as f:
            f.write(s)

        exe = Path(path + 'pnflow.exe')
        subprocess.Popen([exe, file])



# from collections import ChainMap  # Might use eventually
import numpy as np
from chemicals import Vm_to_rho, rho_to_Vm
from chemicals import numba_vectorized
from chemicals import R, k
from openpnm.phases import GenericPhase as GenericPhase
from openpnm.utils import HealthDict, PrintableList, SubDict
from openpnm.utils import Docorator, GenericSettings
from openpnm import models
from openpnm.utils import logging
logger = logging.getLogger(__name__)
docstr = Docorator()


class GenericMixtureSettings(GenericSettings):
    r"""
    The following settings are specific to Mixture objects

    Parameters
    ----------
    components : list of strings
        The names of each pure component object that constitute the mixture

    """

    components = []
    prefix = 'mix'


@docstr.get_sections(base='GenericMixture', sections=['Parameters'])
@docstr.dedent
class GenericMixture(GenericPhase):
    r"""
    Creates Phase object that represents a multicomponent mixture system
    consisting of a given list of OpenPNM Phase objects as components.

    Parameters
    ----------
    %(GenericPhase.parameters)s
    components : list of OpenPNM Phase objects
        A list of all components that constitute this mixture

    Notes
    -----
    All mixtures assume that mole fractions are always stored as
    ``'pore.mole_fraction'`` and that concentration is always stored as
    ``'pore.concentration'``.

    """

    def __init__(self, components=[], settings={}, **kwargs):
        self.settings._update_settings_and_docs(GenericMixtureSettings)
        self.settings.update(settings)
        super().__init__(settings={'prefix': 'mix'}, **kwargs)

        # Add any supplied phases to the phases list
        for comp in components:
            self.set_component(comp)
            self.set_mole_fraction(comp, np.nan)
        self['pore.mole_fraction.all'] = np.nan

        logger.warning('Mixtures are a beta feature and functionality may '
                       + 'change in future versions')

    def __getitem__(self, key):
        try:
            vals = super().__getitem__(key)
        except KeyError:
            try:
                # If key ends in component name, fetch it
                if key.split('.')[-1] in self.settings['components']:
                    comp = self.project[key.split('.')[-1]]
                    vals = comp[key.rsplit('.', maxsplit=1)[0]]
                    return vals
                # If does not end in component name, see if components have it
                vals = SubDict()
                for comp in self.components.keys():
                    vals[key + '.' + comp] = self.components[comp][key]
            except KeyError:
                vals = self.interleave_data(key)
        return vals

    def __setitem__(self, key, value):
        # Prevent writing 'element.property.component' on mixture
        invalid_keys = set(self.props(deep=True)).difference(set(self.props()))
        if key in invalid_keys:
            raise Exception(key + ' already assigned to a component object')
        # If writing a concentration, use set_concentration setter
        if key.startswith('pore.concentration'):
            self.set_concentration(component=key.split('.')[-1], values=value)
        super().__setitem__(key, value)

    def props(self, deep=False, **kwargs):
        temp = PrintableList()
        if deep:
            for item in self.components.values():
                temp.extend([prop+'.'+item.name for prop in item.props(**kwargs)])
        temp.extend(super().props(**kwargs))
        temp.sort()
        return temp

    def __str__(self):
        horizontal_rule = '―' * 78
        lines = super().__str__()
        lines = '\n'.join((lines, 'Component Phases', horizontal_rule))
        for item in self.components.values():
            lines = '\n'.join((lines, item.__module__.replace('__', '')
                               + ' : ' + item.name))
        lines = '\n'.join((lines, horizontal_rule))
        return lines

    def _update_total_molfrac(self):
        temp = np.zeros((self.Np, ), dtype=float)
        comps = list(self.components.values())
        for item in comps:
            temp += self['pore.mole_fraction.' + item.name]
        self['pore.mole_fraction.all'] = temp

    def _reset_molfracs(self):
        for item in self.settings['components']:
            self['pore.mole_fraction.'+item] = np.nan
        self['pore.mole_fraction.all'] = np.nan

    def update_concentrations(self, mode='all'):
        r"""
        Updates all unspecified concentrations from mole fractions and molar
        density.

        Parameters
        ----------
        mode : str {'all' (default), 'update'}
            If 'all' then all concentrations will be updated based on the
            current mole fractions.  If 'update', then only concentrations
            that are missing or filled with *nans* will be updated.

        Notes
        -----
        This function requires that molar density be defined on the mixture.

        """
        if 'pore.molar_density' not in self.keys():
            raise Exception("Cannot update concentration without "
                            + "\'pore.molar_density\'")
        comps = self.settings['components'].copy()
        for item in comps:
            c = self['pore.mole_fraction.' + item] * self['pore.molar_density']
            if mode == 'all':
                self['pore.concentration.' + item] = c
            elif mode == 'update':
                if 'pore.concentration.' + item not in self.items():
                    self['pore.concentration.' + item] = c

    def update_mole_fractions(self, free_comp=None):
        r"""
        Updates mole fraction values so the sum of all mole fractions is
        1.0 in each pore.

        Parameters
        ----------
        free_comp : OpenPNM object
            Specifies which component's mole fraction should be adjusted to
            enforce the sum of all mole fractions to equal 1.0.  See ``Notes``
            for more details.

        Notes
        -----
        If ``free_comp`` is not given then a search is conducted to find a
        component whose mole fractions are *nans* and this component is
        treated as ``free_comp``.

        If more than one component, or *no* components, is/are found during
        this search, an attempt is made to determine mole fractions from
        the concentrations, such that $x_i = C_i/C_total$.

        """
        # Parse the free_comp argument
        if hasattr(free_comp, 'name'):  # Get name if give openpnm object
            free_comp = free_comp.name
        if free_comp is not None:  # Set the component's mole fraction to nan
            self['pore.mole_fraction.' + free_comp] = np.nan
        # Scan the list of components and find if any have nans
        comps = self.settings['components'].copy()
        hasnans = []
        for item in comps:
            # If component not found, give it nans
            if 'pore.mole_fraction.'+item not in self.keys():
                self['pore.mole_fraction.'+item] = np.nan
            hasnans.append(np.any(np.isnan(self['pore.mole_fraction.' + item])))
        # If only one has nans then update it's mole fraction to make sum = 1.0
        if hasnans.count(True) == 1:
            # Remove free_comp from list
            comp = comps.pop(hasnans.index(True))
            # Set comp with nans to 1.0, then deduct other mole fracs
            self['pore.mole_fraction.'+comp] = 1.0
            for item in comps:
                self['pore.mole_fraction.' + comp] -= \
                    self['pore.mole_fraction.' + item]
        # If all or none of the components have values, then use concentrations
        else:
            # First find total concentration of all components
            total_conc = 0.0
            try:
                for item in comps:
                    total_conc += self['pore.concentration.' + item]
                # Then find mole fractions as C_i/C_total
                for item in comps:
                    mf = self['pore.concentration.' + item] / total_conc
                    self['pore.mole_fraction.' + item] = mf
            except KeyError:
                logger.warning("Insufficient concentrations defined, cannot "
                               + "recalculate mole fractions")
        self._update_total_molfrac()
        self.regenerate_models(['pore.molar_mass'])

    def set_concentration(self, component, values=[]):
        r"""
        Specify the concentration of a component in each pore

        Parameters
        ----------
        component : OpenPNM Phase object or name
            The phase object of the component whose concentration is being
            specified
        values : array_like
            The concentration of the given ``component `` in each pore.  This
            array must be *Np*-long, with one value for each pore in the
            network.  If a scalar is received it is applied to all pores.

        See Also
        --------
        set_mole_fraction

        """
        if isinstance(component, str):
            component = self.components[component]
        Pvals = np.array(values, ndmin=1)
        if component not in self.components.values():
            raise Exception('Given component not part of mixture')
        if Pvals.size:
            key = 'pore.concentration.' + component.name
            super().__setitem__(key, Pvals)

    def set_mole_fraction(self, component, values=[]):
        r"""
        Specify mole fraction of each component in each pore

        Parameters
        ----------
        components : OpenPNM Phase object or name string
            The phase whose mole fraction is being specified
        values : array_like
            The mole fraction of the given ``component `` in each pore.  This
            array must be *Np*-long, with one value between 0 and 1 for each
            pore in the network.  If a scalar is received it is applied to
            all pores.

        """
        # Parse the input args
        if isinstance(component, str):
            component = self.components[component]
        Pvals = np.array(values, ndmin=1)
        # If given component not part of mixture, set it
        if component.name not in self.settings['components']:
            self.set_component(component)
        if np.any(Pvals > 1.0) or np.any(Pvals < 0.0):
            logger.warning('Received values contain mole fractions outside '
                           + 'the range of 0 to 1')
        if Pvals.size:
            self['pore.mole_fraction.' + component.name] = Pvals
        self._update_total_molfrac()

    def _del_comps(self, components):
        if not isinstance(components, list):
            components = [components]
        # Remove from settings:
        for comp in components:
            self.settings['components'].remove(comp.name)
        # Remove data from dict:
        for item in list(self.keys()):
            if item.endswith(comp.name):
                self.pop(item)
        self['pore.mole_fraction.all'] = np.nan

    def _get_comps(self):
        comps = {}
        comps.update({item: self.project[item]
                      for item in sorted(self.settings['components'])})
        return comps

    def _set_comps(self, components):
        if not isinstance(components, list):
            components = [components]
        temp = self.settings['components'].copy()
        temp.extend([val.name for val in components])
        comps = list(set(temp))
        self.settings['components'] = comps
        # Add mole_fraction array to dict, filled with nans
        for item in comps:
            self['pore.mole_fraction.' + item] = np.nan

    components = property(fget=_get_comps, fset=_set_comps)

    def set_component(self, component, mode='add'):
        r"""
        Add another component to the mixture

        Parameters
        ----------
        component : OpenPNM Phase object
            The phase object of the component to add.  Can also be a list
            of phases to add more than one at a time.
        mode : string {'add', 'remove'}
            Indicates whether to add or remove the give component(s)

        """
        if mode == 'add':
            self._set_comps(component)
        if mode == 'remove':
            self._del_comps(component)

    def check_mixture_health(self):
        r"""
        Checks the "health" of the mixture

        Calculates the mole fraction of all species in each pore and returns
        an list of where values are too low or too high

        Returns
        -------
        health : dict
            A HealtDict object containing lists of locations where the mole
            fractions are not unity.  One value indiates locations that are
            too high, and another where they are too low.

        """
        h = HealthDict()
        h['mole_fraction_too_low'] = []
        h['mole_fraction_too_high'] = []
        self._update_total_molfrac()
        lo = np.where(self['pore.mole_fraction.all'] < 1.0)[0]
        hi = np.where(self['pore.mole_fraction.all'] > 1.0)[0]
        if len(lo) > 0:
            h['mole_fraction_too_low'] = lo
        if len(hi) > 0:
            h['mole_fraction_too_high'] = hi
        return h


class LiquidMixture(GenericMixture):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.add_model(propname='pore.molecular_weight',
                       model=mixture_molecular_weight,
                       regen_mode='deferred')
        self.add_model(propname='pore.viscosity',
                       model=liquid_mixture_viscosity,
                       regen_mode='deferred')
        self.add_model(propname='pore.critical_volume',
                       model=liquid_mixture_critical_volume,
                       regen_mode='deferred')
        self.add_model(propname='pore.critical_temperature',
                       model=liquid_mixture_critical_temperature,
                       regen_mode='deferred')
        self.add_model(propname='pore.acentric_factor',
                       model=mixture_acentric_factor,
                       regen_mode='deferred')
        self.add_model(propname='pore.density',
                       model=liquid_mixture_density,
                       regen_mode='deferred')
        self.add_model(propname='pore.thermal_conductivity',
                       model=liquid_mixture_thermal_conductivity,
                       regen_mode='deferred')
        self.add_model(propname='pore.heat_capacity',
                       model=mixture_heat_capacity,
                       regen_mode='deferred')


class GasMixture(GenericMixture):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.add_model(propname='pore.molecular_weight',
                       model=mixture_molecular_weight,
                       regen_mode='deferred')
        self.add_model(propname='pore.gas_viscosity',
                       model=gas_mixture_viscosity,
                       regen_mode='deferred')
        self.add_model(propname='pore.thermal_conductivity',
                       model=gas_mixture_thermal_conductivity,
                       regen_mode='deferred')
        self.add_model(propname='pore.heat_capacity',
                       model=mixture_heat_capacity,
                       regen_mode='deferred')
        self.add_model(propname='pore.LJ_epsilon',
                       model=gas_mixture_LJ_epsilon,
                       regen_mode='deferred')
        self.add_model(propname='pore.LJ_sigma',
                       model=gas_mixture_LJ_sigma,
                       regen_mode='deferred')
        self.add_model(propname='pore.LJ_omega',
                       model=gas_mixture_LJ_collision_integral,
                       regen_mode='deferred')
        self.add_model(propname='pore.diffusion_coefficient',
                       model=gas_mixture_diffusivity,
                       regen_mode='deferred')


def gas_mixture_viscosity(target):
    ys = [target['pore.mole_fraction.' + c.name] for c in target.components.values()]
    MWs = [c.parameters['molecular_weight'] for c in target.components.values()]
    mus = [c['pore.viscosity'] for c in target.components.values()]
    # Should be a one-liner, but not working
    # mu = numba_vectorized.Herning_Zipperer(ys, mus, MWs, sqrtMWs)
    num = 0.0
    denom = 0.0
    for i in range(len(ys)):
        num += ys[i]*mus[i]*np.sqrt(MWs[i])
        denom += ys[i]*np.sqrt(MWs[i])
    mu = num/denom
    return mu


def liquid_mixture_viscosity(target):
    xs = [target['pore.mole_fraction.' + c.name] for c in target.components.values()]
    mus = [c['pore.viscosity'] for c in target.components.values()]
    mu = 0.0
    for i in range(len(xs)):
        mu += xs[i]*np.log(mus[i])
    return np.exp(mu)


def liquid_mixture_critical_volume(target):
    xs = [target['pore.mole_fraction.' + c.name] for c in target.components.values()]
    Vcs = [c.parameters['critical_volume'] for c in target.components.values()]
    N = len(xs)  # Number of components

    Vm1 = np.sum([xs[i]*Vcs[i] for i in range(N)], axis=0)
    Vm2 = np.sum([xs[i]*(Vcs[i])**(2/3) for i in range(N)], axis=0)
    Vm3 = np.sum([xs[i]*(Vcs[i])**(1/3) for i in range(N)], axis=0)
    Vm = 0.25*(Vm1 + 3*(Vm2*Vm3))
    return Vm


def liquid_mixture_critical_temperature(target, Vc='pore.critical_volume'):
    xs = [target['pore.mole_fraction.' + c.name] for c in target.components.values()]
    Tcs = [c.parameters['critical_temperature'] for c in target.components.values()]
    Vcs = [c.parameters['critical_volume'] for c in target.components.values()]
    Vm = target[Vc]
    N = len(xs)  # Number of components

    num = np.zeros_like(xs[0])
    for i in range(N):
        for j in range(N):
            VT = (Vcs[i]*Tcs[i]*Vcs[j]*Tcs[j])**0.5
            num += xs[i]*xs[j]*VT
    Tcm = num/Vm
    return Tcm


def mixture_acentric_factor(target):
    xs = [target['pore.mole_fraction.' + c.name] for c in target.components.values()]
    omegas = [c.parameters['acentric_factor'] for c in target.components.values()]
    omega = np.sum([omegas[i]*xs[i] for i in range(len(xs))], axis=0)
    return omega


def mixture_molecular_weight(target):
    xs = [target['pore.mole_fraction.' + c.name] for c in target.components.values()]
    MWs = [c.parameters['molecular_weight'] for c in target.components.values()]
    MW = np.zeros_like(xs[0])
    for i in range(len(xs)):
        MW += xs[i]*MWs[i]
    return MW


def liquid_mixture_density(target, temperature='pore.temperature'):
    # Actually pure component density using mixture properties
    T = target[temperature]
    MW = target['pore.molecular_weight']
    Tc = target['pore.critical_temperature']
    Vc = target['pore.critical_volume']
    omega = target['pore.acentric_factor']
    T = np.clip(T, -np.inf, Tc)
    Tr = T/Tc
    tau = 1.0 - Tr
    tau_cbrt = (tau)**(1.0/3.)
    a = 0.296123
    b = 0.0480645
    c = 0.0427258
    d = 0.386914
    e = 0.190454
    f = 0.81446
    g = 1.43907
    h = 1.52816
    V_delta = (-a + Tr*(Tr*(-b*Tr - c) + d))/(Tr - 1.00001)
    V_0 = tau_cbrt*(tau_cbrt*(tau_cbrt*(e*tau_cbrt - f) + g) - h) + 1.0
    Vm = Vc*V_0*(1.0 - omega*V_delta)
    rhoL = Vm_to_rho(Vm=Vm, MW=MW)
    return rhoL


def liquid_mixture_thermal_conductivity(target, density='pore.density'):
    xs = [target['pore.mole_fraction.' + c.name] for c in target.components.values()]
    kLs = [c['pore.thermal_conductivity'] for c in target.components.values()]
    # kL = numba_vectorized.DIPPR9I(xs, kLs)  # Another one that doesn't work
    Vm = [rho_to_Vm(c[density], c.parameters['molecular_weight'])
          for c in target.components.values()]
    denom = np.sum([xs[i]*Vm[i] for i in range(len(xs))], axis=0)
    phis = np.array([xs[i]*Vm[i] for i in range(len(xs))])/denom
    kij = 2/np.sum([1/kLs[i] for i in range(len(xs))], axis=0)
    kmix = np.zeros_like(xs[0])
    N = len(xs)
    for i in range(N):
        for j in range(N):
            kmix += phis[i]*phis[j]*kij
    return kmix


def gas_mixture_thermal_conductivity(target, temperature='pore.temperature'):
    T = target[temperature]
    ys = [target['pore.mole_fraction.' + c.name] for c in target.components.values()]
    kGs = [c['pore.thermal_conductivity'] for c in target.components.values()]
    MWs = [c.parameters['molecular_weight'] for c in target.components.values()]
    # numba_vectorized.Lindsay_Bromley(T, ys, kGs, mus, Tbs, MWs)
    kmix = np.zeros_like(T)
    for i in range(len(ys)):
        num = ys[i]*kGs[i]
        A = 0.0
        denom = 0.0
        for j in range(len(ys)):
            A += np.sqrt(MWs[j]/MWs[i])
            denom += ys[j]*A
        kmix += num/denom
    return kmix


def mixture_heat_capacity(target):
    zs = [target['pore.mole_fraction.' + c.name] for c in target.components.values()]
    Cps = [c['pore.heat_capacity'] for c in target.components.values()]
    Cpmix = np.sum([zs[i]*Cps[i] for i in range(len(zs))], axis=0)
    return Cpmix


def gas_mixture_LJ_collision_integral(
        target,
        temperature='pore.temperature',
        epsilon='pore.LJ_epsilon',
        ):
    T = target[temperature]
    eAB = target[epsilon]
    A = 1.06036
    B = 0.15610
    C = 0.19300
    D = 0.47635
    E = 1.03587
    F = 1.52996
    G = 1.76474
    H = 3.89411
    Ts = k*T/eAB
    Omega = A/Ts**B + C/np.exp(D*Ts) + E/np.exp(F*Ts) + G/np.exp(H*Ts)
    return Omega


def gas_mixture_LJ_epsilon(target):
    es = [c.parameters['lennard_jones_epsilon'] for c in target.components.values()]
    eAB = np.sqrt(np.prod(es))
    return eAB


def gas_mixture_LJ_sigma(target):
    ss = [c.parameters['lennard_jones_sigma'] for c in target.components.values()]
    sAB = np.mean(ss)
    return sAB


def gas_mixture_diffusivity(
        target,
        temperature='pore.temperature',
        pressure='pore.pressure',
        sigma='pore.LJ_sigma',
        omega='pore.LJ_omega',
        ):
    MW = [c.parameters['molecular_weight'] for c in target.components.values()]
    MWAB = 2/np.sum(1/np.array(MW))
    T = target[temperature]
    P = target[pressure]
    sAB = target[sigma]
    Omega = target[omega]
    DAB = 0.00266*T*1.5/(P*MWAB**0.5*sAB**2*Omega)*101.325  # Convert to m2/s
    return DAB



from openpnm.phases.mixtures import GenericSpecies
from openpnm.utils import PrintableDict
import chemicals as chem
from chemicals import numba_vectorized
from chemicals import Vm_to_rho
from chemicals.utils import R, k
from openpnm.utils import logging
import numpy as np
logger = logging.getLogger(__name__)


class SpeciesByName(GenericSpecies):
    r"""
    Creates Phase object that represents a single species in a multicomponent
    mixture system.

    Parameters
    ----------
    network : OpenPNM Network object
        The network to which this phase object will be attached.
    species : str
        The name of the species to generate.  This is used to lookup tabulated
        constants in the ``chemicals`` package, which attempts to find a match. For
        instance, 'water', 'Water', and 'H2O' all work.
    project : OpenPNM Project object, optional
        The Project with which this phase should be associted.  If a
        ``network`` is given then this is ignored and the Network's project
        is used.  If a ``network`` is not given then this is mandatory.
    name : string, optional
        The name of the phase.  This is useful to keep track of the objects
        throughout the simulation.  The name must be unique to the project.
        If no name is given, one is generated.

    """

    def __init__(self, species, **kwargs):
        super().__init__(**kwargs)
        CAS = chem.CAS_from_any(species)
        self.settings['CAS'] = CAS
        self.parameters = PrintableDict()
        self.parameters['molecular_weight'] = chem.MW(CAS)
        self.parameters['critical_temperature'] = chem.critical.Tc(CAS)
        self.parameters['critical_pressure'] = chem.critical.Pc(CAS)
        self.parameters['critical_volume'] = chem.critical.Vc(CAS)
        self.parameters['critical_compressibilty_factor'] = chem.critical.Zc(CAS)
        self.parameters['boiling_temperature'] = chem.Tb(CAS)
        self.parameters['melting_temperature'] = chem.Tm(CAS)
        self.parameters['acentric_factor'] = chem.acentric.omega(CAS)
        self.parameters['dipole_moment'] = chem.dipole.dipole_moment(CAS)
        if CAS in extra_LJ.keys():
            s, e_k = extra_LJ[CAS]
        else:
            e_k = chem.lennard_jones.Stockmayer(CAS)
            s = chem.lennard_jones.molecular_diameter(CAS)
        if e_k is not None:
            self.parameters['lennard_jones_epsilon'] = e_k*k
        if s is not None:
            self.parameters['lennard_jones_sigma'] = s


class GasByName(SpeciesByName):
    def __init__(self, species, **kwargs):
        super().__init__(species=species, **kwargs)
        self.add_model(propname='pore.heat_capacity',
                       model=gas_heat_capacity)
        self.add_model(propname='pore.thermal_conductivity',
                       model=gas_thermal_conductivity)
        self.add_model(propname='pore.viscosity',
                       model=gas_viscosity)


class LiquidByName(SpeciesByName):
    def __init__(self, species, **kwargs):
        super().__init__(species=species, **kwargs)
        self.add_model(propname='pore.gas_heat_capacity',
                       model=gas_heat_capacity)
        self.add_model(propname='pore.heat_capacity',
                       model=liquid_heat_capacity)
        self.add_model(propname='pore.thermal_conductivity',
                       model=liquid_thermal_conductivity)
        self.add_model(propname='pore.viscosity',
                       model=liquid_viscosity)
        self.add_model(propname='pore.density',
                       model=liquid_density)
        self.add_model(propname='pore.vapor_pressure',
                       model=vapor_pressure)


extra_LJ = {}
extra_LJ['7727-37-9'] = (3.788, 71.4)


def vapor_pressure(target, temperature='pore.temperature'):
    T = target[temperature]
    CAS = target.settings['CAS']
    Tc = target.parameters['critical_temperature']
    try:
        coeffs = chem.vapor_pressure.Psat_data_AntoineExtended.loc[CAS]
        _, A, B, C, Tc, to, n, E, F, Tmin, Tmax = coeffs
        PV = numba_vectorized.TRC_Antoine_extended(T, A, B, C, n, E, F)
    except KeyError:
        coeffs = chem.vapor_pressure.Psat_data_AntoinePoling.loc[CAS]
        _, A, B, C, Tmin, Tmax = coeffs
        PV = 10**(A - B/(T + C))
        # PV = numba_vectorized.vapor_pressure.Antoine(T=T, A=A, B=B, C=C)
    return PV


def liquid_density(target, temperature='pore.temperature'):
    T = target[temperature]
    MW = target.parameters['molecular_weight']
    Tc = target.parameters['critical_temperature']
    Vc = target.parameters['critical_volume']
    omega = target.parameters['acentric_factor']
    Vm = numba_vectorized.COSTALD(T, Tc, Vc, omega)
    rhoL = Vm_to_rho(Vm, MW)
    return rhoL


def liquid_viscosity(target, temperature='pore.temperature'):
    T = target[temperature]
    MW = target.parameters['molecular_weight']
    Tc = target.parameters['critical_temperature']
    Pc = target.parameters['critical_pressure']
    omega = target.parameters['acentric_factor']
    muL = numba_vectorized.Letsou_Stiel(T, MW, Tc, Pc, omega)
    return muL


def gas_viscosity(target, temperature='pore.temperature'):
    T = target[temperature]
    MW = target.parameters['molecular_weight']
    Tc = target.parameters['critical_temperature']
    Pc = target.parameters['critical_pressure']
    muG = numba_vectorized.viscosity_gas_Gharagheizi(T, Tc, Pc, MW)
    return muG


def liquid_thermal_conductivity(target, temperature='pore.temperature'):
    T = target[temperature]
    MW = target.parameters['molecular_weight']
    Tb = target.parameters['boiling_temperature']
    Pc = target.parameters['critical_pressure']
    omega = target.parameters['acentric_factor']
    kL = numba_vectorized.Gharagheizi_liquid(T, MW, Tb, Pc, omega)
    return kL


def gas_thermal_conductivity(target, temperature='pore.temperature'):
    T = target[temperature]
    MW = target.parameters['molecular_weight']
    Tb = target.parameters['boiling_temperature']
    Pc = target.parameters['critical_pressure']
    omega = target.parameters['acentric_factor']
    kG = numba_vectorized.Gharagheizi_gas(T, MW, Tb, Pc, omega)
    return kG


def liquid_heat_capacity(target,
                         Cpgm='pore.gas_heat_capacity',
                         T='pore.temperature'):
    T = target['pore.temperature']
    Cpgm = target['pore.gas_heat_capacity']
    Tc = target.parameters['critical_temperature']
    omega = target.parameters['acentric_factor']
    Cplm = numba_vectorized.Rowlinson_Poling(T, Tc, omega, Cpgm)
    return Cplm


def gas_heat_capacity(target, temperature='pore.temperature'):
    T = target[temperature]
    props = chem.heat_capacity.TRC_gas_data.loc[target.settings['CAS']]
    _, Tmin, Tmax, a0, a1, a2, a3, a4, a5, a6, a7, I, J, Hfg = props
    Cp = numba_vectorized.TRCCp(T, a0, a1, a2, a3, a4, a5, a6, a7)
    return Cp



r"""
The hydraulic size factor is the geometrical part of the pre-factor in
Stoke's flow:

.. math::

    Q = \frac{A^2}{8 \mu L} \Delta P
      = \frac{S_{hydraulic}}{\mu} \Delta P

Thus :math:`S_{hydraulic}` represents the combined effect of the area
and length of the *conduit*, which consists of a throat and 1/2 of the
pores on each end.

"""

import numpy as _np
import openpnm.models.geometry.conduit_lengths as _conduit_lengths
import openpnm.geometry.GenericGeometry as _GenericGeometry
from .misc import _get_conduit_diameters

__all__ = [
    "spheres_and_cylinders",
    "circles_and_rectangles",
    "cones_and_cylinders",
    "pyramids_and_cuboids",
    "cubes_and_cuboids",
    "squares_and_rectangles",
    "intersecting_cones",
    "intersecting_pyramids",
    "ncylinders_in_series"
]


def spheres_and_cylinders(
    target: _GenericGeometry,
    pore_diameter="pore.diameter",
    throat_diameter="throat.diameter",
):
    r"""
    Computes hydraulic size factors for conduits assuming pores are
    spheres and throats are cylinders.

    Parameters
    ----------
    target : GenericGeometry
        Geometry object which this model is associated with. This controls
        the length of the calculated array, and also provides access to
        other necessary properties.
    pore_diameter : str
        Dictionary key of the pore diameter values.
    throat_diameter : str
        Dictionary key of the throat diameter values.

    Returns
    -------
    dict
        Dictionary containing conduit size factors. Size factors are
        accessible via the keys: 'pore1', 'throat', and 'pore2'.

    Notes
    -----
    The hydraulic size factor is the geometrical part of the pre-factor in
    Stoke's flow:

    .. math::

        Q = \frac{A^2}{8 \mu L} \Delta P
          = \frac{S_{hydraulic}}{\mu} \Delta P

    Thus :math:`S_{hydraulic}` represents the combined effect of the area
    and length of the *conduit*, which consists of a throat and 1/2 of the
    pores on each end.

    """
    D1, Dt, D2 = _get_conduit_diameters(target, pore_diameter, throat_diameter)
    L1, Lt, L2 = _conduit_lengths.spheres_and_cylinders(
        target, pore_diameter=pore_diameter, throat_diameter=throat_diameter
    ).T

    # Fi is the integral of (1/A^2) dx, x = [0, Li]
    a = 4 / (D1**3 * _np.pi**2)
    b = 2 * D1 * L1 / (D1**2 - 4 * L1**2) + _np.arctanh(2 * L1 / D1)
    F1 = a * b
    a = 4 / (D2**3 * _np.pi**2)
    b = 2 * D2 * L2 / (D2**2 - 4 * L2**2) + _np.arctanh(2 * L2 / D2)
    F2 = a * b
    Ft = Lt / (_np.pi / 4 * Dt**2)**2

    # I is the integral of (y^2 + z^2) dA, divided by A^2
    I1 = I2 = It = 1 / (2 * _np.pi)

    # S is 1 / (16 * pi^2 * I * F)
    S1 = 1 / (16 * _np.pi**2 * I1 * F1)
    St = 1 / (16 * _np.pi**2 * It * Ft)
    S2 = 1 / (16 * _np.pi**2 * I2 * F2)

    return {"pore1": S1, "throat": St, "pore2": S2}


def circles_and_rectangles(
    target: _GenericGeometry,
    pore_diameter="pore.diameter",
    throat_diameter="throat.diameter",
):
    r"""
    Computes hydraulic size factors for conduits assuming pores are
    circles and throats are rectangles.

    Parameters
    ----------
    target : GenericGeometry
        Geometry object which this model is associated with. This controls
        the length of the calculated array, and also provides access to
        other necessary properties.
    pore_diameter : str
        Dictionary key of the pore diameter values.
    throat_diameter : str
        Dictionary key of the throat diameter values.

    Returns
    -------
    dict
        Dictionary containing conduit size factors. Size factors are
        accessible via the keys: 'pore1', 'throat', and 'pore2'.

    Notes
    -----
    The hydraulic size factor is the geometrical part of the pre-factor in
    Stoke's flow:

    .. math::

        Q = \frac{A^2}{8 \mu L} \Delta P
          = \frac{S_{hydraulic}}{\mu} \Delta P

    Thus :math:`S_{hydraulic}` represents the combined effect of the area
    and length of the *conduit*, which consists of a throat and 1/2 of the
    pores on each end.

    This model should only be used for true 2D networks, i.e. with planar
    symmetry.

    """
    D1, Dt, D2 = _get_conduit_diameters(target, pore_diameter, throat_diameter)
    L1, Lt, L2 = _conduit_lengths.circles_and_rectangles(
        target, pore_diameter=pore_diameter, throat_diameter=throat_diameter
    ).T

    # Fi is the integral of (1/A^3) dx, x = [0, Li]
    F1 = L1 / (D1**2 * _np.sqrt(D1**2 - 4 * L1**2))
    F2 = L2 / (D2**2 * _np.sqrt(D2**2 - 4 * L2**2))
    Ft = Lt / Dt**3

    # S is 1 / (12 * F)
    S1, St, S2 = [1 / (Fi * 12) for Fi in [F1, Ft, F2]]

    return {"pore1": S1, "throat": St, "pore2": S2}


def cones_and_cylinders(
    target: _GenericGeometry,
    pore_diameter="pore.diameter",
    throat_diameter="throat.diameter",
):
    r"""
    Computes hydraulic size factors for conduits assuming pores are
    truncated cones and throats are cylinders.

    Parameters
    ----------
    target : OpenPNM Object
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.
    pore_diameter : str
        Dictionary key of the pore diameter values.
    throat_diameter : str
        Dictionary key of the throat diameter values.

    Returns
    -------
    dict
        Dictionary containing conduit size factors. Size factors are
        accessible via the keys: 'pore1', 'throat', and 'pore2'.

    Notes
    -----
    The hydraulic size factor is the geometrical part of the pre-factor in
    Stoke's flow:

    .. math::

        Q = \frac{A^2}{8 \mu L} \Delta P
          = \frac{S_{hydraulic}}{\mu} \Delta P

    Thus :math:`S_{hydraulic}` represents the combined effect of the area
    and length of the *conduit*, which consists of a throat and 1/2 of the
    pores on each end.

    """
    D1, Dt, D2 = _get_conduit_diameters(target, pore_diameter, throat_diameter)
    L1, Lt, L2 = _conduit_lengths.cones_and_cylinders(
        target, pore_diameter=pore_diameter, throat_diameter=throat_diameter
    ).T

    # Fi is the integral of (1/A^2) dx, x = [0, Li]
    F1 = 16 / 3 * (L1 * (D1**2 + D1 * Dt + Dt**2) / (D1**3 * Dt**3 * _np.pi**2))
    F2 = 16 / 3 * (L2 * (D2**2 + D2 * Dt + Dt**2) / (D2**3 * Dt**3 * _np.pi**2))
    Ft = Lt / (_np.pi * Dt**2 / 4) ** 2

    # I is the integral of (y^2 + z^2) dA, divided by A^2
    I1 = I2 = It = 1 / (2 * _np.pi)

    # S is 1 / (16 * pi^2 * I * F)
    S1 = 1 / (16 * _np.pi**2 * I1 * F1)
    St = 1 / (16 * _np.pi**2 * It * Ft)
    S2 = 1 / (16 * _np.pi**2 * I2 * F2)

    return {"pore1": S1, "throat": St, "pore2": S2}


def trapezoids_and_rectangles(
    target: _GenericGeometry,
    pore_diameter="pore.diameter",
    throat_diameter="throat.diameter",
):
    r"""
    Computes hydraulic size factors for conduits assuming pores are
    trapezoids and throats are rectangles.

    Parameters
    ----------
    target : OpenPNM Object
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.
    pore_diameter : str
        Dictionary key of the pore diameter values.
    throat_diameter : str
        Dictionary key of the throat diameter values.

    Returns
    -------
    dict
        Dictionary containing conduit size factors. Size factors are
        accessible via the keys: 'pore1', 'throat', and 'pore2'.

    Notes
    -----
    The hydraulic size factor is the geometrical part of the pre-factor in
    Stoke's flow:

    .. math::

        Q = \frac{A^2}{8 \mu L} \Delta P
          = \frac{S_{hydraulic}}{\mu} \Delta P

    Thus :math:`S_{hydraulic}` represents the combined effect of the area
    and length of the *conduit*, which consists of a throat and 1/2 of the
    pores on each end.

    This model should only be used for true 2D networks, i.e. with planar
    symmetry.

    """
    D1, Dt, D2 = _get_conduit_diameters(target, pore_diameter, throat_diameter)
    L1, Lt, L2 = _conduit_lengths.cones_and_cylinders(
        target, pore_diameter=pore_diameter, throat_diameter=throat_diameter
    ).T

    # Fi is the integral of (1/A^3) dx, x = [0, Li]
    F1 = L1 / 2 * (D1 + Dt) / (D1 * Dt)**2
    F2 = L2 / 2 * (D2 + Dt) / (D2 * Dt)**2
    Ft = Lt / Dt**3

    # S is 1 / (12 * F)
    S1, St, S2 = [1 / (Fi * 12) for Fi in [F1, Ft, F2]]

    return {"pore1": S1, "throat": St, "pore2": S2}


def pyramids_and_cuboids(
    target,
    pore_diameter="pore.diameter",
    throat_diameter="throat.diameter",
):
    r"""
    Computes hydraulic size factors for conduits assuming pores are
    truncated pyramids and throats are cuboids.

    Parameters
    ----------
    target : GenericGeometry
        Geometry object which this model is associated with. This controls
        the length of the calculated array, and also provides access to
        other necessary properties.
    pore_diameter : str
        Dictionary key pointing to the pore diameter values.
    throat_diameter : str
        Dictionary key pointing to the throat diameter values.

    Returns
    -------
    dict
        Dictionary containing conduit size factors. Size factors are
        accessible via the keys: 'pore1', 'throat', and 'pore2'.

    Notes
    -----
    The hydraulic size factor is the geometrical part of the pre-factor in
    Stoke's flow:

    .. math::

        Q = \frac{A^2}{8 \mu L} \Delta P
          = \frac{S_{hydraulic}}{\mu} \Delta P

    Thus :math:`S_{hydraulic}` represents the combined effect of the area
    and length of the *conduit*, which consists of a throat and 1/2 of the
    pores on each end.

    """
    D1, Dt, D2 = _get_conduit_diameters(target, pore_diameter, throat_diameter)
    L1, Lt, L2 = _conduit_lengths.pyramids_and_cuboids(
        target, pore_diameter=pore_diameter, throat_diameter=throat_diameter
    ).T

    # Fi is the integral of (1/A^2) dx, x = [0, Li]
    F1 = 1 / 3 * (L1 * (D1**2 + D1 * Dt + Dt**2) / (D1**3 * Dt**3))
    F2 = 1 / 3 * (L2 * (D2**2 + D2 * Dt + Dt**2) / (D2**3 * Dt**3))
    Ft = Lt / Dt**4

    # I is the integral of (y^2 + z^2) dA, divided by A^2
    I1 = I2 = It = 1 / 6

    # S is 1 / (16 * pi^2 * I * F)
    S1 = 1 / (16 * _np.pi**2 * I1 * F1)
    St = 1 / (16 * _np.pi**2 * It * Ft)
    S2 = 1 / (16 * _np.pi**2 * I2 * F2)

    return {"pore1": S1, "throat": St, "pore2": S2}


def cubes_and_cuboids(
    target: _GenericGeometry,
    pore_diameter="pore.diameter",
    throat_diameter="throat.diameter",
    pore_aspect=[1, 1, 1],
    throat_aspect=[1, 1, 1],
):
    r"""
    Computes hydraulic size factors for conduits assuming pores are cubes
    and throats are cuboids.

    Parameters
    ----------
    target : GenericGeometry
        Geometry object which this model is associated with. This controls
        the length of the calculated array, and also provides access to
        other necessary properties.
    pore_diameter : str
        Dictionary key pointing to the pore diameter values.
    throat_diameter : str
        Dictionary key pointing to the throat diameter values.
    pore_aspect : list
        Aspect ratio of the pores
    throat_aspect : list
        Aspect ratio of the throats

    Returns
    -------
    dict
        Dictionary containing conduit size factors. Size factors are
        accessible via the keys: 'pore1', 'throat', and 'pore2'.

    Notes
    -----
    The hydraulic size factor is the geometrical part of the pre-factor in
    Stoke's flow:

    .. math::

        Q = \frac{A^2}{8 \mu L} \Delta P
          = \frac{S_{hydraulic}}{\mu} \Delta P

    Thus :math:`S_{hydraulic}` represents the combined effect of the area
    and length of the *conduit*, which consists of a throat and 1/2 of the
    pores on each end.

    """
    D1, Dt, D2 = _get_conduit_diameters(target, pore_diameter, throat_diameter)
    L1, Lt, L2 = _conduit_lengths.cubes_and_cuboids(
        target, pore_diameter=pore_diameter, throat_diameter=throat_diameter
    ).T

    # Fi is the integral of (1/A^2) dx, x = [0, Li]
    F1 = L1 / D1**4
    F2 = L2 / D2**4
    Ft = Lt / Dt**4

    # I is the integral of (y^2 + z^2) dA, divided by A^2
    I1 = I2 = It = 1 / 6

    # S is 1 / (16 * pi^2 * I * F)
    S1 = 1 / (16 * _np.pi**2 * I1 * F1)
    St = 1 / (16 * _np.pi**2 * It * Ft)
    S2 = 1 / (16 * _np.pi**2 * I2 * F2)

    return {"pore1": S1, "throat": St, "pore2": S2}


def squares_and_rectangles(
    target: _GenericGeometry,
    pore_diameter="pore.diameter",
    throat_diameter="throat.diameter",
    pore_aspect=[1, 1],
    throat_aspect=[1, 1],
):
    r"""
    Computes hydraulic size factors for conduits assuming pores are
    squares and throats are rectangles.

    Parameters
    ----------
    target : GenericGeometry
        Geometry object which this model is associated with. This controls
        the length of the calculated array, and also provides access to
        other necessary properties.
    pore_diameter : str
        Dictionary key pointing to the pore diameter values.
    throat_diameter : str
        Dictionary key pointing to the throat diameter values.
    pore_aspect : list
        Aspect ratio of the pores
    throat_aspect : list
        Aspect ratio of the throats

    Returns
    -------
    dict
        Dictionary containing conduit size factors. Size factors are
        accessible via the keys: 'pore1', 'throat', and 'pore2'.

    Notes
    -----
    The hydraulic size factor is the geometrical part of the pre-factor in
    Stoke's flow:

    .. math::

        Q = \frac{A^2}{8 \mu L} \Delta P
          = \frac{S_{hydraulic}}{\mu} \Delta P

    Thus :math:`S_{hydraulic}` represents the combined effect of the area
    and length of the *conduit*, which consists of a throat and 1/2 of the
    pores on each end.

    This model should only be used for true 2D networks, i.e. with planar
    symmetry.

    """
    D1, Dt, D2 = _get_conduit_diameters(target, pore_diameter, throat_diameter)
    L1, Lt, L2 = _conduit_lengths.squares_and_rectangles(
        target, pore_diameter=pore_diameter, throat_diameter=throat_diameter
    ).T

    # Fi is the integral of (1/A^3) dx, x = [0, Li]
    F1 = L1 / D1**3
    F2 = L2 / D2**3
    Ft = Lt / Dt**3

    # S is 1 / (12 * F)
    S1, St, S2 = [1 / (Fi * 12) for Fi in [F1, Ft, F2]]

    return {"pore1": S1, "throat": St, "pore2": S2}


def intersecting_cones(
    target: _GenericGeometry,
    pore_diameter="pore.diameter",
    throat_diameter="throat.diameter",
    midpoint=None,
):
    r"""
    Computes hydraulic size factors for conduits of intersecting cones.

    Parameters
    ----------
    target : GenericGeometry
        Geometry object which this model is associated with. This controls
        the length of the calculated array, and also provides access to
        other necessary properties.
    pore_diameter : str
        Dictionary key pointing to the pore diameter values.
    throat_diameter : str
        Dictionary key pointing to the throat diameter values.
    midpoint : str, optional
        Dictionary key of the midpoint values.

    Returns
    -------
    dict
        Dictionary containing conduit size factors. Size factors are
        accessible via the keys: 'pore1', 'throat', and 'pore2'.

    Notes
    -----
    The hydraulic size factor is the geometrical part of the pre-factor in
    Stoke's flow:

    .. math::

        Q = \frac{A^2}{8 \mu L} \Delta P
          = \frac{S_{hydraulic}}{\mu} \Delta P

    Thus :math:`S_{hydraulic}` represents the combined effect of the area
    and length of the *conduit*, which consists of a throat and 1/2 of the
    pores on each end.

    """
    raise NotImplementedError


def intersecting_trapezoids(
    target: _GenericGeometry,
    pore_diameter="pore.diameter",
    throat_diameter="throat.diameter",
    midpoint=None,
):
    r"""
    Computes hydraulic size factors for conduits of intersecting
    trapezoids.

    Parameters
    ----------
    target : GenericGeometry
        Geometry object which this model is associated with. This controls
        the length of the calculated array, and also provides access to
        other necessary properties.
    pore_diameter : str
        Dictionary key pointing to the pore diameter values.
    throat_diameter : str
        Dictionary key pointing to the throat diameter values.
    midpoint : str, optional
        Dictionary key of the midpoint values.

    Returns
    -------
    dict
        Dictionary containing conduit size factors. Size factors are
        accessible via the keys: 'pore1', 'throat', and 'pore2'.

    Notes
    -----
    The hydraulic size factor is the geometrical part of the pre-factor in
    Stoke's flow:

    .. math::

        Q = \frac{A^2}{8 \mu L} \Delta P
          = \frac{S_{hydraulic}}{\mu} \Delta P

    Thus :math:`S_{hydraulic}` represents the combined effect of the area
    and length of the *conduit*, which consists of a throat and 1/2 of the
    pores on each end.

    This model should only be used for true 2D networks, i.e. with planar
    symmetry.

    """
    raise NotImplementedError


def intersecting_pyramids(
    target: _GenericGeometry,
    pore_diameter="pore.diameter",
    throat_diameter="throat.diameter",
    midpoint=None,
):
    r"""
    Computes hydraulic size factors for conduits of intersecting pyramids.

    Parameters
    ----------
    target : GenericGeometry
        Geometry object which this model is associated with. This controls
        the length of the calculated array, and also provides access to
        other necessary properties.
    pore_diameter : str
        Dictionary key pointing to the pore diameter values.
    throat_diameter : str
        Dictionary key pointing to the throat diameter values.
    midpoint : str, optional
        Dictionary key of the midpoint values.

    Returns
    -------
    dict
        Dictionary containing conduit size factors. Size factors are
        accessible via the keys: 'pore1', 'throat', and 'pore2'.

    Notes
    -----
    The hydraulic size factor is the geometrical part of the pre-factor in
    Stoke's flow:

    .. math::

        Q = \frac{A^2}{8 \mu L} \Delta P
          = \frac{S_{hydraulic}}{\mu} \Delta P

    Thus :math:`S_{hydraulic}` represents the combined effect of the area
    and length of the *conduit*, which consists of a throat and 1/2 of the
    pores on each end.

    """
    raise NotImplementedError


def ncylinders_in_series(
    target: _GenericGeometry,
    pore_diameter="pore.diameter",
    throat_diameter="throat.diameter",
    n=5,
):
    r"""
    Computes hydraulic size factors for conduits of spheres and cylinders
    with the spheres approximated as N cylinders in series.

    Parameters
    ----------
    target : GenericGeometry
        Geometry object which this model is associated with. This controls
        the length of the calculated array, and also provides access to
        other necessary properties.
    pore_diameter : str
        Dictionary key pointing to the pore diameter values.
    throat_diameter : str
        Dictionary key pointing to the throat diameter values.
    n : int
        Number of cylindrical divisions for each pore

    Returns
    -------
    dict
        Dictionary containing conduit size factors. Size factors are
        accessible via the keys: 'pore1', 'throat', and 'pore2'.

    Notes
    -----
    The hydraulic size factor is the geometrical part of the pre-factor in
    Stoke's flow:

    .. math::

        Q = \frac{A^2}{8 \mu L} \Delta P
          = \frac{S_{hydraulic}}{\mu} \Delta P

    Thus :math:`S_{hydraulic}` represents the combined effect of the area
    and length of the *conduit*, which consists of a throat and 1/2 of the
    pores on each end.

    """
    D1, Dt, D2 = _get_conduit_diameters(target, pore_diameter, throat_diameter)
    # Ensure throats are never bigger than connected pores
    Dt = _np.minimum(Dt, 0.99 * _np.minimum(D1, D2))
    L1, Lt, L2 = _conduit_lengths.spheres_and_cylinders(
        target, pore_diameter=pore_diameter, throat_diameter=throat_diameter
    ).T
    dL1 = _np.linspace(0, L1, num=n)
    dL2 = _np.linspace(0, L2, num=n)
    r1 = D1 / 2 * _np.sin(_np.arccos(dL1 / (D1 / 2)))
    r2 = D2 / 2 * _np.sin(_np.arccos(dL2 / (D2 / 2)))

    gtemp = (_np.pi * r1 ** 4 / (8 * L1 / n)).T
    F1 = 1 / _np.sum(1 / gtemp, axis=1)
    gtemp = (_np.pi * r2 ** 4 / (8 * L2 / n)).T
    F2 = 1 / _np.sum(1 / gtemp, axis=1)
    Ft = (_np.pi * (Dt / 2) ** 4 / (8 * Lt)).T

    return {"pore1": F1, "throat": Ft, "pore2": F2}



import openpnm.models as mods
import openpnm.models.geometry as gmods
from openpnm.geometry import GenericGeometry


class SquaresAndRectangles(GenericGeometry):
    r"""
    Squares and Rectangles subclass of GenericGeometry. This subclass is
    meant as a basic default geometry to get started quickly.

    Pore diameters are randomly assigned between 0 and the largest sphere
    that does not overlap with it's nearest neighbor.

    Throat diameters are half the diameter of the smaller of it's two
    neighboring pores.

    Parameters
    ----------
    network : GenericNetwork
        The network with which this Geometry should be associated
    project : Project, optional
        Can be supplied instead of a ``network``
    pores : array_like
        The pores in the domain where this Geometry applies
    throats : array_like
        The throats in the domain where this Geometry applies
    name : str
        The name of the object, which is also used as the label where this
        geometry is defined.

    Examples
    --------
    Geometry objects (along with Physics objects) can be applied to a
    subset of pores and/or throats.  This allows for different geometrical
    property models to be applied in different regions. This is illustrated
    in the following code:

    >>> import numpy as np
    >>> import scipy as sp
    >>> import openpnm as op
    >>> import matplotlib.pyplot as plt
    >>> from openpnm.geometry import SquaresAndRectangles
    >>> pn = op.network.CubicDual(shape=[5, 5, 5])
    >>> Ps = pn.pores('primary')
    >>> Ts = pn.throats('primary')
    >>> geo1 = SquaresAndRectangles(network=pn, pores=Ps, throats=Ts)
    >>> Ps = pn.pores('secondary')
    >>> Ts = pn.throats(['secondary', 'interconnect'])
    >>> geo2 = SquaresAndRectangles(network=pn, pores=Ps, throats=Ts)

    Now override the 'pore.diameter' values on the ``geo2`` object:

    >>> geo2.remove_model('pore.diameter')  # Remove model and data
    >>> geo2['pore.diameter'] = np.random.rand(geo2.Np) * 0.05

    Look at the 'pore.diameter' distributions on each object:

    >>> fig = plt.hist(geo1['pore.diameter'], bins=20, alpha=0.5)
    >>> fig = plt.hist(geo2['pore.diameter'], bins=20, alpha=0.5)

    The resulting figure shows that these two Geometry object each have a
    different pore size distribution, with ``geo2`` being much smaller:

    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.add_model(propname='pore.seed',
                       model=mods.misc.random,
                       element='pore',
                       num_range=[0.2, 0.7],
                       seed=None)

        self.add_model(propname='pore.max_size',
                       model=mods.geometry.pore_size.largest_sphere,
                       iters=10)

        self.add_model(propname='pore.diameter',
                       model=mods.misc.product,
                       prop1='pore.max_size',
                       prop2='pore.seed')

        self.add_model(propname='pore.volume',
                       model=mods.geometry.pore_volume.square,
                       pore_diameter='pore.diameter')

        self.add_model(propname='throat.max_size',
                       model=mods.misc.from_neighbor_pores,
                       mode='min',
                       prop='pore.diameter')

        self.add_model(propname='throat.diameter',
                       model=mods.misc.scaled,
                       factor=0.5,
                       prop='throat.max_size')

        self.add_model(propname='throat.length',
                       model=mods.geometry.throat_length.squares_and_rectangles,
                       pore_diameter='pore.diameter',
                       throat_diameter='throat.diameter')

        self.add_model(propname='throat.cross_sectional_area',
                       model=mods.geometry.throat_cross_sectional_area.rectangle,
                       throat_diameter='throat.diameter')

        self.add_model(propname='throat.volume',
                       model=mods.geometry.throat_volume.rectangle,
                       throat_diameter='throat.diameter',
                       throat_length='throat.length')

        self.add_model(propname='throat.diffusive_size_factors',
                       model=gmods.diffusive_size_factors.squares_and_rectangles,
                       pore_diameter="pore.diameter",
                       throat_diameter="throat.diameter")

        self.add_model(propname='throat.hydraulic_size_factors',
                       model=gmods.hydraulic_size_factors.squares_and_rectangles,
                       pore_diameter="pore.diameter",
                       throat_diameter="throat.diameter")



r"""
Helper methods for openpnm.geometry module.

"""


def _get_conduit_diameters(target, pore_diameter, throat_diameter):
    r"""
    Private helper methods for fetching conduit diameters.
    """
    network = target.project.network
    throats = target.throats(target=network)
    cn = network["throat.conns"][throats]

    D1 = network[pore_diameter][cn[:, 0]]
    Dt = network[throat_diameter][throats]
    D2 = network[pore_diameter][cn[:, 1]]

    return D1, Dt, D2
